// SP70: a program for unsupervised learning in the ICMAUS framework.

// Copyright (C) 2003 J G Wolff.

// This program is released to the public domain without any restrictions.
// You can redistribute it and/or modify it as you wish.

// In any publication that refers to any version of this program, please
// acknowledge J G Wolff as the originator of the program. 

// A description of this program may be found in "Unsupervised learning
// in a framework of information compression by multiple alignment,
// unification and search", J G Wolff, March 2002. A copy may be obtained
// from http://uk.arxiv.org/abs/cs.AI/0302015. See also
// http://www.cognitionresearch.org.uk/papers/ul/ul.htm.

// Please retain this notice on all versions.

// Dr J G Wolff, www.CognitionResearch.org.uk.
// Email: jgw@cognitionresearch.org.uk.

// May 2003.

/*
** TITLE:		SP70_head.h
**
** VERSION: 	SP70, v 9.2
**
** ABSTRACT:	The header file to be inherited by all modules in
**				the SP system. It includes the class structure for the
**				program.
**
** ENVIRONMENT: This file should be placed on the `include' path of
**				the C++ compiler so that it may be located correctly.
**				This file produces no executable code.
**
** AUTHOR:		Gerry Wolff
**
** MODIFIED BY:
**
** SP70:
**
** 2.0 JGW 7/4/99 - Derived from SP61, v 4.5.
** 2.1	Backup.
** 2.2 JGW 10/8/99 -	See comp.cpp.
** 2.3 JGW 14/9/99 -	See comp.cpp.
** 2.4 JGW 1/10/99 -	See comp.cpp.
** 2.5 JGW 10/11/99 -	See comp.cpp.
** 2.6 JGW 11/11/99 -	See comp.cpp.
** 2.7 JGW 13/11/99 -	See comp.cpp.
** 2.8 JGW 18/11/99 -	See comp.cpp.
** 3.0 JGW 22/2/00 -	See comp.cpp.
** 3.1 JGW 25/2/00 -	See comp.cpp.
** 3.2 JGW 8/5/00 - 	See comp.cpp.
** 3.3 JGW 16/5/00 -	See comp.cpp.
** 3.4 JGW 26/5/00 -	See comp.cpp.
** 3.5 JGW 5/6/00 - 	See comp.cpp.
** 3.6 JGW 7/7/00 - 	See comp.cpp.
** 3.7 JGW 11/7/00 -	See comp.cpp.
** 3.8 JGW 1/8/00 - 	See comp.cpp.
** 4.0 JGW 29/9/00 -	See SP70_comp.cpp.
** 4.1 JGW 5/10/00 -	See SP70_comp.cpp.
** 4.2 JGW 13/10/00 -	See SP70_comp.cpp.
** 4.3 JGW 28/10/00 -	See SP70_comp.cpp.
** 4.4 JGW 30/10/00 -	See SP70_comp.cpp.
** 4.5 JGW 13/11/00 -	See SP70_comp.cpp.
** 4.6 JGW 2/1/01 -		See SP70_comp.cpp.
** 4.7 JGW 11/1/01 -	See SP70_comp.cpp.
** 4.8 JGW 13/2/01 -	See SP70_comp.cpp.
** 4.9 JGW 22/2/01 -	See SP70_comp.cpp.
** 5.0 JGW 8/3/01 -		See SP70_comp.cpp.
** 5.1 JGW 4/4/01 -		See SP70_comp.cpp.
** 5.2 JGW 18/4/01 -	See SP70_comp.cpp.
** 5.3 JGW 18/4/01 -	See SP70_comp.cpp.
** 5.4 JGW 27/4/01 -	See SP70_comp.cpp.
** 5.5 JGW 12/6/01 -	See SP70_comp.cpp.
** 6.0 JGW 26/6/01 -	See SP70_comp.cpp.
** 6.1 JGW 6/7/01 -		See SP70_comp.cpp.
** 6.2 JGW 16/7/01 -	See SP70_comp.cpp.
** 6.3 JGW 17/7/01 -	See SP70_comp.cpp.
** 6.4 JGW 17/7/01 -	See SP70_comp.cpp.
** 6.5 JGW 19/7/01 -	See SP70_comp.cpp.
** 7.0 JGW 24/7/01 -	See SP70_comp.cpp.
** 7.1 JGW 30/7/01 -	See SP70_comp.cpp.
** 7.2 JGW 2/8/01 -		See SP70_comp.cpp.
** 7.3 JGW 8/8/01 -		See SP70_comp.cpp.
** 9.1 JGW 6/11/01 -	See SP70_comp.cpp.
** 9.2 JGW 9/11/01 -	See SP70_comp.cpp.
*/

#include <iostream.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

// Definitions.

#define BEST_ALIGNMENT_PRINT_LIMIT 5
#define BASIC_PATTERN 1
#define ABSTRACT_PATTERN 2
#define CODE_PATTERN 3
#define SUB_ALIGNMENT 4
#define COMPOSITE_ALIGNMENT 5
#define ALIGNMENT_FROM_PARSING 6
#define OUTPUT_COUNTER_1 5
#define OUTPUT_COUNTER_2 10
#define GAP_BETWEEN_AL_COLS 3
#define H 1
#define V 2
#define HIT_NODE_ARRAY_SIZE 10000
#define CODE_BIT_COST 8.0
#define BRACKET_BIT_COST 4.0
#define VERY_LARGE_NUMBER 1000000000
#define ALIGNMENTS_FOR_LEARNING 3 // Controls the
	// number of alignments fed into 'learning' at the end
	// of the 'recognition' phase for each window from New.
#define PRINTING_POINTS_PER_MM 2.8452756
#define CHARACTER_HEIGHT_WIDTH_RATIO 1.95
#define FONT_SET_SIZE 10
#define PRINTING_SPACE_LONG_SIDE 247 // The length (in mm) of the long
	// side of the space available for printing on an A4 sheet.
#define PRINTING_SPACE_SHORT_SIDE 160 // The length (in mm) of the short
	// side of the space available for printing on an A4 sheet.
#define LANDSCAPE_CHARS_PER_LINE 116 // Characters (12 pt) in one line of
	// an A4 page in landscape orientation (25 mm margins).
#define LANDSCAPE_LINES_PER_PAGE 21 // Lines in one A4 page in
	// landscape orientation (12 pt characters, 25 mm margins).
#define KEY_ARRAY_SIZE 1000
#define ELLIPSIS_MARGIN 5 // write_al_chars_limit must be at least
	// ELLIPSIS_MARGIN smaller than WRITE_ALIGNMENT_CHAR_LENGTH
	// to allow space for inserting "..." at the end of truncated
	// alignments.
#define HIGH_VALUE 1000000000 // An arbitarily high integer value used
	// in unify_best and elsewhere.
#define LOW_VALUE -1000000000 // An arbitarily low integer value used
	// in unify_best and tree_object::sort_children().
#define CONTEXT_CODE_SYMBOL 1 // Value for the type of a symbol.
#define DISCRIMINATION_SYMBOL 2 // Value for the type of a symbol.
#define DATA_SYMBOL 3 // Value for the type of a symbol.
#define LEFT_BRACKET 4 // Value for the type of a symbol.
#define RIGHT_BRACKET 5 // Value for the type of a symbol.
#define IDENTIFICATION 1 // Value for the status of a symbol.
	// See sp70_od, %47.
#define CONTENTS 2 // Value for the status of a symbol.
	// See sp70_od, %47.
#define ORIGINAL 1
#define COPY 2
#define MAX_ALIGNMENT_DEPTH 50
#define MAX_NEW_ARRAY_ROWS 200
#define SCALING_FACTORS_SIZE 100
#define SMALL_SCRATCH_ARRAY_SIZE 10
#define MEDIUM_SCRATCH_ARRAY_SIZE 100
#define LARGE_SCRATCH_ARRAY_SIZE 1000
#define AL_CONTENTS_ARRAY_WIDTH 50
#define WRITE_ALIGNMENT_C_COLS 20
#define MCOLS 200 /* The number of columns in the best_hit_structure[] */
#define START 0
#define CONTINUE 1
#define SEQUENCE 1
#define GROUP 2
#define ON 1
#define OFF 0
#define LETTERS 2
#define DIGITS 3
#define LANDSCAPE 0
#define PORTRAIT 1
#define ALREADY_VISITED_1 1
#define ALREADY_VISITED_2 2
#define UNKNOWN 2
#define BUFSIZE 5000
#define NIL 0
#define NULL_VALUE -1
#define STRING_LENGTH 20
#define LONG_STRING_LENGTH 100
#define ESCAPE 033
#define DEL '\177'
#define PRINT_ALL_FREQUENCIES 3
#define PRINT_GROUP_FREQUENCIES 2
#define PRINT_SEQUENCE_FREQUENCIES 1
#define PRINT_NO_FREQUENCIES 0
#define MAX_WRITE_LINE 75

// Macro definitions

#define list_for(X, Y, Z) for (X = (Y *)Z->get_first_child();X != NIL ; \
	X = (Y *)Z->get_next_child())

// Definitions of classes and declarations of methods.

// CLASS STRUCTURE

//	base_element
//		list_element
//		alignment_element

//	base_object
//		symbol
//		combination
//		grammar
//		class_record
//		tree_object
//			group
//			cost_tree_object
//				sequence
//					sub_alignment
//				hit_node

class base_element ;
class list_element ;
class alignment_element ;

class base_object ;
class symbol ;
class combination ;
class grammar ;
class class_record ;
class tree_object ;
class group ;
class cost_tree_object ;
class sequence ;
class sub_alignment ;
class hit_node ;

struct key_entry
{
	char *abbreviated_symbol ;
	char *complete_symbol ;
} ;

struct row_entry
{
	sequence *row_pattern ;
	list_element *patt_last_el_pos ; // This is used to record the
		// last (ie, most recent) el_pos for row_pattern where the
		// corresponding symbol has been written into
		// write_al_chars[][].
	int fe_col_in_write_al_chars ; // Keeps track of the first empty
		// column ***for this row*** in write_al_chars[][].
	bool pattern_is_finished ; // true when all the symbols in
		// the row_pattern have been written, false otherwise.
} ;

struct sub_alignment_entry
{
	sub_alignment *sub_al1 ;
	int pos_new_first, pos_new_last ; // Records the position
		// in the current pattern from New of the first
		// and last hit symbols of sub_alignment
		// in that pattern.
} ;

struct hn_entry
{
	hit_node *new_hit_node ;
	hit_node *hn_parent ;
} ;

// Definitions of global variables

extern bool right_bracket_expected ;
extern group *set_of_class_records ;
extern int number_of_patterns_in_new ;
extern int number_of_patterns_in_old ;
extern int number_of_symbols_in_new ;
extern int number_of_symbols_in_old ;
extern double best_score_this_window ;
extern int fe_hit_node_array ;
extern struct hn_entry hit_node_array[] ;
extern sequence *receptacle_pattern ;
extern list_element *cns_el_pos ;
extern int half_hit_structure_rows ;
extern int total_number_of_alignments ; 
extern sequence *symbol_selection_array[][MCOLS] ;
extern double scaling_factors[] ;
extern clock_t time1 ;
extern clock_t time2 ; 
extern int cycle ; /* Counts the recognition cycles within one window. */
extern int window ; /* Counts the windows which have been processed. */
extern int phase ;
extern int new_pattern_number ;
extern int original_alphabet_size ;
extern int current_alphabet_size ;
extern char *in_filename ;
extern char *parameters_filename ;
extern char *out_filename ;
extern char *latex_filename ;
extern char *plot_filename ;
extern FILE *input_file ;
extern FILE *parameters_file ;
extern FILE *output_file ;
extern FILE *latex_file ;
extern FILE *plot_file ;
extern int code_number ;
extern bool identification_symbols_marked ;
extern hit_node **half_list ;
extern char **write_al_chars ;
extern char **write_al_vacant_slots ;
extern double font_heights[] ;
extern int fe_key_array ;
extern struct key_entry key_array[] ;
extern struct row_entry row_array[] ;
extern const int max_alignment_depth ; /* The maximum depth of an sequence
	that can be written (symbols). */
extern const int max_alignment_length ;
extern int write_al_chars_length ;
extern int write_section_chars_length ;
extern double minimum_font_height ;
extern double maximum_font_height ;
extern int maximum_learning_count ;
extern int show_all_parsing_alignments ;
extern char alignment_format ;
extern int hit_structure_rows ;
extern int patterns_to_keep ;
extern double cost_factor ;
extern hit_node **leaf_array ; /* leaf_array[] is used for recording the
	leaf nodes of the hit structure. */
extern int *sort_array ; /* sort_array[] is used for sorting the
	entries in leaf_array[]. */
extern hit_node *hit_root ;
extern hit_node *hn_master ;
extern int fe_sort ; /* The first empty entry in sort_array[]. */
extern group *corpus ;
extern group *old_patterns ;
extern group *symbols_in_new ;
extern group *new_patterns ;
extern group *set_of_sequences ;
extern sequence *current_new_pattern ;
extern group *original_symbols_in_corpus ;
extern group *symbols_in_old_after_learning ;
extern group *driving_patterns_and_alignments ;
extern group *keep_alignments ;
extern group *current_alignments_from_parsing ;
extern group *full_alignments_from_parsing ;
extern group *global_alignments_from_parsing ;
extern group *created_alignments ;
extern group *created_code_patterns ;
extern group *set_of_grammars ;
extern group *set_of_cleaned_up_grammars ;
extern group *copies_of_grammar_patterns ;
extern grammar *naive_grammar ;
extern class_record *current_class_record ;
extern bool probabilities ;
extern bool show_al_structure ;
extern bool show_h_structure ;
extern bool show_unselected_alignments ;
extern bool verbose ;
extern int use_abbreviations ;
extern int keep_rows ;
extern int driving_keep_rows ;
extern int full_alignment_limit ;
extern int grammar_limit ;
extern int extraction_limit ;
extern int window_size ;
extern int fail_score ;
extern char figure_ID[] ;
extern int max_new_gap ;
extern int max_old_gap ;
extern int max_driving_gap ;
extern int max_target_gap ;
extern int max_unsupported_cycles ;
extern int max_alignments_in_one_cycle ;
extern int orientation ;
extern int group_ID_number ;
extern int sequence_ID_number ;
extern int hit_node_ID_number ;
extern int combination_ID_number ;
extern int grammar_ID_number ;
extern int class_record_ID_number ;
extern sequence **alignments_array ;

/* Declarations of global functions. */

extern symbol *make_bracket_symbol(int, int, int) ;
extern bool create_patterns_and_sort() ;
extern void purge_lists() ;
extern void increment_index(int *, int, char *) ;
extern class_record *find_class_record_by_name(char *) ;
extern class_record *find_class_record_from_pattern(sequence *) ;
extern void sort_and_print_class_records() ;
extern sequence *recognise_select_and_write() ;
extern void compile_class_records() ;
extern void probabilities_of_inferences(sequence *) ;
extern void write_new_and_old() ;
extern void assign_symbol_frequencies_and_costs(group *) ;
extern void find_symbol_frequencies(group *, bool, bool) ;
extern void assign_type_to_symbols() ;
extern void assign_status_to_symbols() ;
extern bool recognise(bool) ;
extern bool find_best_matches(bool) ;
extern void find_symbol_type_information_costs(group *) ;
extern void compute_scaling_factors() ;
extern void modified_sizes_of_sfe_codes(group *) ;
extern bool is_unique(sequence *, symbol *, 
	sequence **) ;
extern void make_selection(bool) ;
extern sequence *unify_best(int, bool, bool) ;
extern void set_up_hit_structure() ;
extern sequence *make_one_alignment(hit_node *) ;
extern sequence *finish_u_b(sequence *) ;
extern bool mismatch_found(hit_node *) ;
extern void prepare_hit_sequence(hit_node *) ;
extern symbol *make_column(int, sequence *, symbol *, 
	int, sequence *, symbol *, int, int, int, bool,
	bool) ;
extern void print_pattern_window_cycle() ;
extern sub_alignment *make_unmatched_sub_alignment(sequence *, 
	list_element *) ;
extern void SP_main() ;
extern void time_now() ;
extern symbol *make_ccs() ;
extern void output_alignment(sequence *) ;
extern void sort_leaf_array(int, int) ;
extern bool is_better_than(double, double) ;
extern void make_letter_abbreviations() ;
extern void make_digit_abbreviations() ;
extern void write_key() ;
extern double log_2(double) ;
extern void plus_one(int *, int, char *) ;
extern void show_hit_structure(hit_node *, int) ;
extern base_object *get_child(list_element **) ;
extern void finish_deletion() ;
extern void exit_routine(int) ;
extern void clear_nodes(hit_node *) ;
extern void read_parameters() ;
extern void write_lines(FILE *, char *, int) ;
extern void abort_run(char *) ;
extern void write_message(char *) ;
extern void load() ;
extern hit_node *get_leaf_nodes_in_order(int *, sequence **, int *) ;

// CLASSES

class base_element
{
	public:
		base_element(base_object *obj1=NIL) {el_obj = obj1 ;}
		base_element(base_element &el) {el_obj = el.el_obj ;}
		~base_element() {}
	
		void set_el_obj(base_object *obj1){el_obj = obj1 ;}
		base_object *get_el_obj() {return(el_obj) ;}
	protected:
		base_object *el_obj ; // Points to a symbol when used
			// within a basic pattern. Points to an object 
			// containing an array of alignment_elements
			// when used within an alignment.
} ; // class base_element

class alignment_element: public base_element
{
	public:
		alignment_element(base_object *obj1=NIL) : base_element(obj1)
			{same_column_above = same_column_below = NULL_VALUE ;
			original_pattern = NIL ;
			orig_patt_int_pos = NULL_VALUE ;}
		alignment_element(alignment_element &al_el1) : 
				base_element(al_el1)
			{same_column_above = al_el1.get_same_column_above() ;
			same_column_below = al_el1.get_same_column_below() ;
			original_pattern = al_el1.get_original_pattern() ;
			orig_patt_int_pos = al_el1.get_orig_patt_int_pos() ;}
		~alignment_element() {} // Notice that
			// any el_obj which attached to the alignment
			// element is *not* deleted.

		void set_same_column_above(int row) {same_column_above = row ;}
		int get_same_column_above() {return(same_column_above) ;}
		void set_same_column_below(int row) {same_column_below = row ;}
		int get_same_column_below() {return(same_column_below) ;}
		void set_original_pattern(sequence *orig_patt)
			{original_pattern = orig_patt ;}
		sequence *get_original_pattern() {return(original_pattern) ;}
		void set_orig_patt_int_pos(int pos)
			{orig_patt_int_pos = pos ;}
		int get_orig_patt_int_pos() {return(orig_patt_int_pos) ;}
		void show_al_structure() ;
	protected:

		// The el_obj field (at the level of base_element) is used
		// to enter the address of the symbol for this al_el.

		int same_column_above ; // An index to to an al_el in a row
			// above which contains a matching symbol.
		int same_column_below ; // An pointer to to an al_el in a row 
			// below which contains a matching symbol.
		sequence *original_pattern ; // The pattern from which
			// this row of the sequence and this al_el
			// was derived.
		int orig_patt_int_pos ; // The position of the attached
			// symbol in the original pattern.

		// Note: the column position of each al_el is the 'position'
		// field of the column in which the al_el appears.
		// The row position of each al_el is the index value
		// for the al_el within the column in which the al_el
		// appears.
} ; // class alignment_element

class list_element: public base_element
{
	public:
		list_element(base_object *obj1=NIL) : base_element(obj1)
			{parent = NIL ; position = NULL_VALUE ; next = NIL ;}
		list_element(list_element &el1) : base_element(el1)
			{parent = NIL ; position = NULL_VALUE ; next = NIL ;}
		~list_element() {}

		bool this_is_last_position()
			{if (next == NIL) return(true) ; else return(false) ;}
		void set_parent(base_object *parent1) {parent = parent1 ;}
		base_object *get_parent() {return(parent) ;}
		void set_position(int pos) {position = pos ;}
		int get_position(){return(position) ;}
		void set_next(list_element *el1){next = el1 ;}
		list_element *get_next(){return(next) ;}
	protected:
		base_object *parent ; // The object containing a pointer
			// to the start of the list.
		int position ; // The position in the list.
		list_element *next ; // The next list_element in the list.
} ; // class list_element

class base_object
{
	public:
		base_object() {frequency = 1 ;}
		base_object(base_object &obj) {frequency = obj.frequency ;}
		virtual ~base_object() {}
		virtual base_object *clone() {return(NIL) ;}

		virtual bool is_symbol() {return(false) ;}
		void increment_frequency() {frequency++ ;}
		void add_to_frequency(int fr) {frequency += fr ;}
		void set_frequency(int fr) {frequency = fr ;}
		int get_frequency() {return(frequency) ;}
		virtual void mark_parent_and_int_positions_recursive() {}
		virtual int count_number_of_symbols() {return(0) ;}
		virtual double total_of_symbol_sizes() {return(0) ;}
		virtual void invert_symbol_status() {}
		virtual void symbol_set_recursion(group *) {}
		virtual void wto(int) {}
		virtual void set_encoding_cost(double) {}
	protected:
		int frequency ;
} ; // class base_object

class symbol: public base_object // This is a class for simple 
	// symbols and for columns of an alignment.
{
	public:
		symbol(char *nm=NIL, int al_depth=1, int hn_ID=NULL_VALUE) ;
		symbol(symbol &s) ;
		~symbol() {delete[] name ;
			delete[] al_el_array ;} // Notice
			// that deletion of al_el_array does *not* delete
			// any el_obj which may be attached to one or
			// alignment elements in the al_el_array.
		base_object *clone() {symbol *result = new symbol(*this) ;
			return(result) ;}
		symbol *shallow_copy() ; // Makes a copy of a symbol 
			// without any information about rows (assumes a depth of 1).

		bool is_symbol() {return(true) ;}
		bool name_matches(symbol *s)
			{if (strcmp(name, s->get_name()) == NIL) 
				return(true) ;
			else return(false) ;}
		void set_name(char *nm) 
			{delete name ; name = new char[strlen(nm) + 1] ;
				strcpy(name, nm) ;}
		char *get_name() {return(name) ;}
		int count_number_of_symbols() {return(1) ;}
		double total_of_symbol_sizes() {return(bit_cost) ;}
		void wto(int) ;
		void set_encoding_cost(double e_cost) {bit_cost = e_cost ;}
		void set_bit_cost(double x) {bit_cost = x ;}
		double get_bit_cost() {return(bit_cost) ;}
		void set_type(int tp) {if (tp != NULL_VALUE
			&& tp != CONTEXT_CODE_SYMBOL && tp != DISCRIMINATION_SYMBOL
			&& tp != DATA_SYMBOL && tp != LEFT_BRACKET
			&& tp != RIGHT_BRACKET)
				abort_run("Invalid type for symbol::set_type") ;
			type = tp ;}
		int get_type() {return(type) ;}
		void set_status(int st) {if (st != NULL_VALUE && st != IDENTIFICATION 
			&& st != CONTENTS) abort_run("Invalid status for symbol::set_type") ;
			status = st ;}
		int get_status() {return(status) ;}
		void mark_parent_and_int_positions_recursive() {}

		// Alignment methods

		bool contains_contents_symbol() ;
		void set_symbol_matches() ;
		symbol *get_row_symbol(int row) {if (sequence_depth == 1 ||
			row > sequence_depth || row < 0)
				abort_run("Invalid value for row in get_row_symbol") ;
			alignment_element *al_el1 = this->get_al_el(row) ;
			return((symbol *)al_el1->get_el_obj()) ;}
		int get_row_orig_int_pos(int row) {if (sequence_depth == 1 ||
			row > sequence_depth || row < 0)
				abort_run("Invalid value for row in get_row_symbol") ;
			alignment_element *al_el1 = this->get_al_el(row) ; 
			return(al_el1->get_orig_patt_int_pos()) ;}
		sequence *get_row_pattern(int row) {if (sequence_depth == 1 ||
			row > sequence_depth || row < 0)
				abort_run("Invalid value for row in get_row_symbol") ;
			alignment_element *al_el1 = this->get_al_el(row) ; 
			return(al_el1->get_original_pattern()) ;}
		void set_row_pattern(int row, sequence *pattern1)
			{if (sequence_depth == 1 ||
				row > sequence_depth || row < 0)
					abort_run("Invalid value for row in set_row_pattern") ;
			alignment_element *al_el1 = this->get_al_el(row) ;
			al_el1->set_original_pattern(pattern1) ;}
		void set_row_symbol(int row, symbol *symbol1)
			{if (sequence_depth == 1 ||
				row > sequence_depth || row < 0)
					abort_run("Invalid value for row in set_row_symbol") ;
			alignment_element *al_el1 = this->get_al_el(row) ;
			al_el1->set_el_obj(symbol1) ;}
		void set_row_orig_int_pos(int row, int orig_int_pos)
			{if (sequence_depth == 1 ||
				row > sequence_depth || row < 0)
					abort_run("Invalid value for row in set_row_orig_int_pos") ;
			alignment_element *al_el1 = this->get_al_el(row) ;
			al_el1->set_orig_patt_int_pos(orig_int_pos) ;}
		void show_al_structure() ;
		alignment_element *get_al_el(int row)
			{if (sequence_depth == 1 ||
				row > sequence_depth || row < 0)
				abort_run("Invalid value for row in get_al_el") ;
			return(al_el_array + row) ;}
		void set_sequence_depth(int al_depth)
			{sequence_depth = al_depth ;}
		int get_sequence_depth() {return(sequence_depth) ;}
		void make_al_el_array(int sequence_depth)
			{al_el_array = new 
				alignment_element[sequence_depth] ;}
		void set_al_el_array(alignment_element *al_el)
			{al_el_array = al_el ;}
		alignment_element *get_al_el_array() {return(al_el_array) ;}
		void set_h_node_ID(int ID)
			{h_node_ID = ID ;}
		int get_h_node_ID() {return(h_node_ID) ;}
		void set_is_a_hit(bool is_hit)
			{symbol_is_a_hit = is_hit ;}
		bool is_a_hit() {return(symbol_is_a_hit) ;}
	protected:
		void symbol_set_recursion(group *set) ;
		char *name ;
		double bit_cost ;
		int type ; // Records whether a given symbol is
			// CONTEXT_CODE_SYMBOL, DISCRIMINATION_SYMBOL, LEFT_BRACKET,
			// RIGHT_BRACKET or DATA_SYMBOL.
		int status ; // Records whether a given symbol is IDENTIFICATION
			// or CONTENTS for the parent pattern.

		// Alignment fields

		int sequence_depth ;
		alignment_element *al_el_array ;
		int h_node_ID ; // The ID number of the hit node (if any)
			// which corresponds to this symbol. The ID is
			// stored rather than the hit node itself so that
			// the hit structure can be properly cleared at the
			// end of each cycle and memory is not cluttered
			// with hit nodes from previous cycles.
		bool symbol_is_a_hit ; // If a column/symbol contains two or
			// more symbols then it is a 'hit' column.
} ; // class symbol

class tree_object: public base_object
{
	public:
		tree_object() : base_object() {first_element = NIL ;
			current_el_pos = NIL ; number_of_children = 0 ;}
		tree_object(tree_object &tn) ;

		void transfer_children(tree_object *recipient)
			{recipient->set_first_element(first_element) ;
				first_element = NIL ;}
		void sort_children() ; // Sorts the children of a node by
			// compression_difference.
		void delete_children() {base_object *child ;
			while (child = this->extract_first_child())
				delete child ;}
		void sort_by_name() ;
		bool is_symbol() {return(false) ;}
		base_object *get_child_by_el_pos(list_element **position)
			{if ((*position) == NIL)
				abort_run("Invalid parameter in get_child_by_el_pos()") ;
				return((*position)->get_el_obj()) ;}
		base_object *get_child_by_int_pos(int position) ;
		base_object *get_next_child() ;
		base_object *get_next_child_by_el_pos(list_element **position) ;
		base_object *get_preceding_child_by_el_pos(list_element **position) ;
		base_object *get_preceding_child(base_object *child) ;
		list_element *get_preceding_el_pos(list_element *position) ;
		list_element *get_next_el_pos(list_element **position)
			{return((*position)->get_next()) ;}
		base_object *get_first_child() {if (first_element == NIL)
			return(NIL) ; current_el_pos = first_element ;
			return(first_element->get_el_obj()) ;}
		base_object *get_last_child() ;
		list_element *get_el_pos_by_int_pos(int position) ;
		list_element *get_el_pos_by_child(base_object *obj) ;
		list_element *get_first_el_pos() {return(first_element) ;}
		list_element *get_last_el_pos()
			{list_element *el_pos1 = first_element, *el_pos2 = NIL ;
			while (el_pos1 != NIL) {el_pos2 = el_pos1 ;
			el_pos1 = el_pos1->get_next() ;} return(el_pos2) ;}
		list_element *get_next_el_pos()
			{if (current_el_pos == NIL) return(first_element) ;
			return(current_el_pos->get_next()) ;}
		void initialise() {current_el_pos = NIL ;}
		void set_current_el_pos(list_element *el_pos)
			{current_el_pos = el_pos ;}
		list_element *get_current_el_pos()
			{return(current_el_pos) ;}
		int get_current_int_pos()
			{if (current_el_pos == NIL) return(NULL_VALUE) ;
			else return(current_el_pos->get_position()) ;}
		int get_int_pos_by_child(base_object *child) ; // Obtain integer
			// position of child from the value stored in its list element.
		int find_int_pos_by_child(base_object *child) ; // Obtain integer
			// position of child by counting.
		void add_child(base_object *child) ;
		void add_child_without_duplicates(base_object *child) ;
		void add_child_at_start(base_object *child) ;
		void extract_child(base_object *child) ;
		base_object *extract_first_child()
			{if (base_object *child = this->get_first_child())
				{this->extract_child(child) ; return(child) ;}
			else return(NIL) ;}
		void precede(base_object *child1, base_object *child2) ;
		void follow(base_object *child1, base_object *child2) ;
		void release_children() ;
		bool this_is_last_child(base_object *child) ;
		bool is_empty() {if (first_element == NIL) return(true) ;
				else return(false) ;}
		symbol *contains_copy_of(symbol *symbol1) ;
		void mark_parent_and_int_positions_recursive() ;
		void mark_parent_and_int_positions_non_recursive() ;
		void set_first_element(list_element *el1)
			{first_element = el1 ;}
		list_element *get_first_element(){return(first_element) ;}
		void write_tree_object(int print_code) ;
		int count_number_of_children() ;
		void increment_child_count() {number_of_children++ ;} 
		void set_number_of_children(int no_of_children)
			{number_of_children = no_of_children ;}
		int get_number_of_children() {return(number_of_children) ;}
		int count_number_of_symbols()
			{int no_of_symbols = 0 ; base_object *child ;
				list_for(child, base_object, this)
					no_of_symbols += child->count_number_of_symbols() ;
				return(no_of_symbols) ;}
		double total_of_symbol_sizes()
			{double cumulative_size = 0 ; base_object *child ;
				list_for(child, base_object, this)
					cumulative_size += child->total_of_symbol_sizes() ;
				return(cumulative_size) ;}
		void invert_symbol_status()
			{base_object *child ; list_for(child, base_object, this)
				child->invert_symbol_status() ;}
		void symbol_set_recursion(group *set) // Called from 
					// symbol_set()
			{base_object *child ; this->initialise() ;
				while (child = this->get_next_child()) 
				child->symbol_set_recursion(set) ;} 
		virtual void print_ID() = 0 ;
		int get_ID() {return(ID) ;}
		group *symbol_set() ;
	protected:
		list_element *first_element ;
		list_element *current_el_pos ;
		int number_of_children ; // Number of children of this node.
		void finish_deletion() ;
		int ID ;
} ; // class tree_object

class group: public tree_object
{
	public:
		group() : tree_object() {ID = group_ID_number++ ;}
		group(group &gr) : tree_object(gr) {ID = group_ID_number++ ;}
		~group() {this->finish_deletion() ;}
		base_object *clone() {group *result = new group(*this) ; 
			return(result) ;}

		void write_patterns(char *heading, int origin) ;
		void write_patterns_with_details() ;
		void print_ID() {fprintf(output_file, "%S%d", "ID", ID) ;}
		void wto(int print_code) ;
		void compile_frequencies(group *symbol_set) ;
		group *compile_alphabet(int *alphabet_size) ;
		void sort_by_ID() ;
		void sort_by_compression_difference() ;
} ; // class group

class cost_tree_object: public tree_object
{
	public:
		cost_tree_object() : tree_object()
			{new_symbols_cost = encoding_cost = compression_ratio = 
				compression_difference = NULL_VALUE ;}
		cost_tree_object(cost_tree_object &cto) : tree_object(cto) 
			{new_symbols_cost = cto.get_new_symbols_cost() ;
				encoding_cost = cto.get_encoding_cost() ;
				compression_ratio = cto.get_compression_ratio() ;
				compression_difference = 
					cto.get_compression_difference() ;}
		~cost_tree_object() {}

		void set_new_symbols_cost(double NSC)
			{new_symbols_cost = NSC ;}
		double get_new_symbols_cost() {return(new_symbols_cost) ;}
		void set_encoding_cost(double EC)
			{encoding_cost = EC ;}
		double get_encoding_cost() {return(encoding_cost) ;}
		void set_compression_ratio(double CR)
			{compression_ratio = CR ;}
		double get_compression_ratio() {return(compression_ratio) ;}
		void set_compression_difference(double CD)
			{compression_difference = CD ;}
		double get_compression_difference() 
			{return(compression_difference) ;}
	protected:
		double new_symbols_cost ;
		double encoding_cost ;
		double compression_ratio ;
		double compression_difference ;
} ; // class cost_tree_object

class sequence: public cost_tree_object
{
	public:
		sequence(int org1=NULL_VALUE) : cost_tree_object()
			{ID = sequence_ID_number++ ; sequence_depth = 1 ; 
			leaf_node_ID = NULL_VALUE ; abs_P = NULL_VALUE ;
			total_cost = 0 ; origin = org1 ;}
		sequence(sequence &seq)  : cost_tree_object(seq)
			{ID = sequence_ID_number++ ;
			sequence_depth = seq.get_sequence_depth() ;
			leaf_node_ID = seq.get_leaf_node_ID() ;
			abs_P = seq.get_abs_P() ; total_cost = seq.get_total_cost() ;
			origin = seq.get_origin() ;}
		base_object *clone() {sequence *result = new sequence(*this) ; 
			return(result) ;}
		~sequence() {this->finish_deletion() ;}

		// Original sequence methods

		void make_or_update_class_records() ;
		symbol *make_discrimination_symbol() ;
		symbol *get_first_ccs()
			{symbol *symbol1 = (symbol *)get_first_child() ;
				if (symbol1->get_type() != LEFT_BRACKET) return(NIL) ;
				symbol1 = (symbol *)get_next_child() ;
				if (symbol1->get_type() != CONTEXT_CODE_SYMBOL)
					return(NIL) ;
				else return(symbol1) ;}
		symbol *get_next_ccs()
			{symbol *symbol1 = (symbol *)get_next_child() ;
				if (symbol1->get_type() != CONTEXT_CODE_SYMBOL)
					return(NIL) ;
				else return(symbol1) ;}
		symbol *get_discrimination_symbol()
			{symbol *symbol1 ;
				list_for(symbol1, symbol, this)
					if (symbol1->get_type() == DISCRIMINATION_SYMBOL)
						return(symbol1) ;
				return(NIL) ;}
		bool has_brackets()
			{list_element *pos1 = NIL ;
				symbol *symbol1 = (symbol *)
					get_next_child_by_el_pos(&pos1) ;
				if (symbol1 == NIL) return(false) ;
				else if (symbol1->get_type() == LEFT_BRACKET
					&& symbol1->get_status() == IDENTIFICATION)
						return(true) ;
				else return(false) ;}
		bool is_full_alignment() ;
		bool all_new_CONTENTS_symbols_matched() ;
		class_record *add_ID_symbols(class_record *class1) ;
		sequence *check_patterns() ; // Checks 'this' against
			// pre-existing patterns to find one containing all
			// and only the same CONTENTS symbols.
		void print_ID() {fprintf(output_file, "%s%d", "ID", ID) ;}
		void wto(int print_code) ;
		bool contents_symbols_match(sequence *pattern1) ;
		void compute_costs() ;
		void set_total_cost(double cost) {total_cost = cost ;}
		double get_total_cost() {return(total_cost) ;}
		void set_origin(int org1) {if (org1 != ABSTRACT_PATTERN &&
			org1 != BASIC_PATTERN && org1 != CODE_PATTERN
			&& org1 != COMPOSITE_ALIGNMENT && org1 != SUB_ALIGNMENT)
				abort_run("Invalid parameter for sequence::set_origin()") ;
			origin = org1 ;}
		int get_origin() {return(origin) ;}

		// Alignment methods

		void correct_column_values() ;
		void compute_score_with_gaps() ;
		void write_out_fully(char *header, sequence *code,
			hit_node *h_node, int write_section_chars_length, 
			int sections_per_page, bool flat_pattern_and_details) ;
		void write_alignment(FILE *of, int wsl, int spp,
				char alignment_format)
			{if (alignment_format == 'H') 
				write_alignment_horizontal(of, wsl, spp) ;
				else write_alignment_vertical(of) ;}
		void write_alignment_vertical(FILE *) ;
		void write_alignment_horizontal(FILE *, int, int) ;
		void write_with_details() ;
		sequence *make_code(bool return_code) ;
		int compute_character_length() ;
		bool alignment_matches(sequence *al) ;
		sequence *matches_earlier_alignment() ;
		bool sequence_matches(sequence *seq1) ;
		bool matches_sub_alignment(sequence *, int, int, int, 
			list_element *, list_element *) ;
		void set_sequence_depth(int al_depth)
			{sequence_depth = al_depth ;}
		int get_sequence_depth() {return(sequence_depth) ;}
		void set_leaf_node_ID(int ID)
			{leaf_node_ID = ID ;}
		int get_leaf_node_ID() {return(leaf_node_ID) ;}
		void set_abs_P(double a_P) {abs_P = a_P ;}
		double get_abs_P() {return(abs_P) ;}
		void show_al_structure(hit_node *h_node) ;
		sequence *get_row_pattern(int row)
			{if (sequence_depth == 1 ||
				row > sequence_depth || row < 0)
				abort_run("Invalid value for row in get_al_el") ;
			symbol *al_col1 = (symbol *)this->get_first_child() ;
			alignment_element *al_el1 = al_col1->get_al_el(row) ;
			return(al_el1->get_original_pattern()) ;}
	protected:
		// Alignment fields
	
		int sequence_depth ; // Records the number of rows in
			// the sequence. An 'alignment' comprising one
			// pattern has a depth of 1.
		int leaf_node_ID ; // The ID number of the leaf hit node
			// (if any) which corresponds to this sequence.
			// The ID is stored rather than the hit node itself
			// so that the hit structure can be properly cleared
			// at the end of each cycle and memory is not cluttered
			// with hit nodes from previous cycles.
		double abs_P ; // The "absolute probability" of an
			// alignment calculated from its EC (in
			// compute_score_with_gaps).
		double total_cost ; // Total of the bit_cost values for all
			// the symbols in the pattern.
		int origin ; // Used to mark the origin of a sequence as 'code',
			// sub-alignment, basic pattern etc.
} ; // class sequence

class sub_alignment: public sequence
{
	public:
		sub_alignment(int org1) : sequence(org1)
			{class_reference = NIL ;}
		sub_alignment(sub_alignment &sub_al1)  : sequence(sub_al1)
			{class_reference = sub_al1.get_class_reference() ;}
		sub_alignment(sequence &seq) : sequence(seq)
			{class_reference = NIL ;}
		~sub_alignment() {}

		void set_class_reference (class_record *ref)
			{class_reference = ref ;}
		class_record *get_class_reference() {return(class_reference) ;}
	protected:
		class_record *class_reference ;
} ; // class sub_alignment

class hit_node: public cost_tree_object	// An element of the hit structure.
{
	public:
		hit_node() ;
		hit_node(hit_node &hn) ;
		~hit_node() {this->finish_deletion() ;}
		base_object *clone() {hit_node *result = new hit_node(*this) ; 
			return(result) ;}

		void print_ID() {fprintf(output_file, "%c%d", '#', ID) ;}
		void set_hn_parent(hit_node *parent) {hn_parent = parent ;}
		hit_node *get_hn_parent() {return(hn_parent) ;}
		void set_driving_pattern(sequence *dr_sequence)
			{driving_pattern = dr_sequence ;}
		sequence *get_driving_pattern() {return(driving_pattern) ;}
		void set_driving_symbol(symbol *dr_symbol)
			{driving_symbol = dr_symbol ;}
		symbol *get_driving_symbol() {return(driving_symbol) ;}
		void set_driving_int_pos(int pos)
			{driving_int_pos = pos ;}
		int get_driving_int_pos() {return(driving_int_pos) ;}
		void set_target_pattern(sequence *al)
			{target_pattern = al ;}
		sequence *get_target_pattern() {return(target_pattern) ;}
		void set_target_symbol(symbol *symb)
			{target_symbol = symb ;}
		symbol *get_target_symbol() {return(target_symbol) ;}
		void set_target_int_pos(int int_pos)
			{target_int_pos = int_pos ;}
		int get_target_int_pos() {return(target_int_pos) ;}
		void set_leaf_entry(hit_node **node) {leaf_entry = node ;}
		hit_node **get_leaf_entry() {return(leaf_entry) ;}
		void wto(int)
			{fprintf(output_file, "hit_node ") ;
			this->print_ID(), fprintf(output_file, "\n\n") ;}
	protected:
		hit_node *hn_parent ;
		sequence *driving_pattern ;
		symbol *driving_symbol ;
		int driving_int_pos ;
		sequence *target_pattern ;
		symbol *target_symbol ;
		int target_int_pos ;
		hit_node **leaf_entry ; // This reverse pointer to an
			// entry in leaf_array[] is needed so that, when
			// a leaf node receives a child (so that it becomes a
			// parent and is no longer a leaf node), the child
			// may use the entry in leaf_array[] that was
			// used by its parent (the old leaf node).
} ; // class hit_node

class combination: public base_object
{
	public:
		combination(int lfs) : base_object()
			{C_ID = combination_ID_number++ ; len_fit_seq = lfs ;
			if (len_fit_seq >= MEDIUM_SCRATCH_ARRAY_SIZE)
				abort_run("len_fit_seq too long in combination") ;
			int i ; for (i = 0; i < len_fit_seq; i++)
				fitting_sequence[i] = false ;
			sub_alignment_list = new group ; 
			combination_score = 0 ;}
		combination(combination &comb1) ;
		~combination(){sub_alignment_list->release_children() ;
			delete sub_alignment_list ;}

		bool is_symbol() {return(false) ;}
		bool is_subset_of(combination *test_comb) ;
		bool can_accept(int pos_new_first, int pos_new_last) ;
		void augment_combination() ;
		void add_sub_alignment(sub_alignment *bal1,
			int bal1_pos_new_first, int bal1_pos_new_last) ;
		group *get_sub_alignment_list()
			{return(sub_alignment_list) ;}
		double get_combination_score() {return(combination_score) ;}
		void set_combination_score(double c_score)
			{combination_score = c_score ;}
		void print_sub_alignments() ;
		void make_composite_alignment() ;
		bool *get_fitting_sequence() {return(fitting_sequence) ;}
		void set_len_fit_seq(int length) {len_fit_seq = length ;}
		int get_len_fit_seq() {return(len_fit_seq) ;}
		void set_C_ID(int cid) {C_ID = cid ;}
		int get_C_ID() {return(C_ID) ;}
	protected:
		bool fitting_sequence[MEDIUM_SCRATCH_ARRAY_SIZE] ; 
			// An array used to mark which symbols 
			// in the current symbol from New
			// are 'covered' by one of the alignments in
			// 'combination' (true) and which are not (false).
		int len_fit_seq ;
		group *sub_alignment_list ; // A list of sub_alignments that
			// fully or partially 'cover' the symbols in New. Each
			// basic alignment contains a pointer to a class record
			// which determines the class symbol to be used when
			// the abstract pattern is built in
			// make_composite_alignment().
		double combination_score ;
		int C_ID ;
} ; // class combination

class grammar: public base_object
{
	public:
		grammar() : base_object() 
			{G = E = score = 0; list_of_patterns = new group ;
			ID = grammar_ID_number++ ;}
		grammar(grammar &gr) ;
		~grammar() {list_of_patterns->release_children() ;}

		void tidy_up_code_symbols() ;
		void remove_unused_context_code_symbols() ;
		void add_pattern(sequence *pattern)
			{list_of_patterns->
				add_child_without_duplicates(pattern) ;}
		void release_patterns() {list_of_patterns->release_children() ;}
		void write_grammar() ;
		void compile_grammar(sequence *full_alignment) ;
		void compute_G() ;
		void set_G(double GG) {G = GG ;}
		double get_G() {return(G) ;}
		void set_E(double EE) {E = EE ;}
		double get_E() {return(E) ;}
		void increment_E(double increment) {E += increment ;}
		void increment_G(double increment) {G += increment ;}
		void set_score(double score1) {score = score1 ;}
		void sum_G_and_E() {score = G + E ;}
		double get_score() {return(score) ;}
		int get_ID() {return(ID) ;}
		group *get_list_of_patterns() {return(list_of_patterns) ;}
		void initialise() {list_of_patterns->initialise() ;}
		sequence *get_next_pattern() 
			{return((sequence *)list_of_patterns->get_next_child()) ;}
		void sort_by_ID() {list_of_patterns->sort_by_ID() ;}
	protected:
		double G ; // Size of grammar in bits.
		double E ; // Size of New after it has been encoded in
			// terms of the grammar.
		double score ; // The sum of G and E.
		int ID ;
		group *list_of_patterns ;
} ; // class grammar

class class_record: public base_object // class_record holds details
	// of each class created by the program. A class is any set
	// of patterns that start with the same context code symbol.
	// A class may contain only one pattern.
{
	public:
		class_record(char *ccsn=NIL) : base_object()
			{ccs_name = NIL ;
				set_ccs_name(ccsn) ;
				ID = class_record_ID_number++ ;
				list_of_patterns = new group ;
				list_of_alignments = new group ;
				set_of_class_records->add_child(this) ;}
		~class_record() {list_of_patterns->release_children();
			delete list_of_patterns ;}

		int get_ID() {return(ID) ;}
		void add_pattern(sequence *pattern1)
			{list_of_patterns->add_child_without_duplicates(pattern1) ;}
		void add_alignment(sequence *al1)
			{list_of_alignments->add_child_without_duplicates(al1) ;}
		void set_ccs_name(char *nm) 
			{delete ccs_name ;
				ccs_name = new char[strlen(nm) + 1] ;
				strcpy(ccs_name, nm) ;}
		char *get_ccs_name()
			{return(ccs_name) ;}
		void write_class_record() ;
		bool symbol_matches_ccs_name(symbol *symbol1)
			{if (strcmp(symbol1->get_name(),
				ccs_name) == 0) return(true) ;
			else return(false) ;}
		void initialise_patterns() {list_of_patterns->initialise() ;}
		sequence *get_next_pattern()
			{return((sequence *)list_of_patterns->get_next_child()) ;}
		void initialise_alignments() {list_of_alignments->initialise() ;}
		sequence *get_next_alignment()
			{return((sequence *)list_of_alignments->get_next_child()) ;}
		void release_alignments()
			{list_of_alignments->release_children() ;}
	protected:
		int ID ;
		char *ccs_name ;
		group *list_of_patterns ; // Contains a list of
			// patterns in this class.
		group *list_of_alignments ; // Contains a list of alignments
			// in this class.
} ; // class class_record
